from rdflib import Graph, Namespace, URIRef

JELLYT = Namespace("https://w3id.org/jelly/dev/tests/vocab#")
MF = Namespace("http://www.w3.org/2001/sw/DataAccess/tests/test-manifest#")


def categorize_by_requires(graph: Graph, test_uri: URIRef) -> str:
    reqs = set(graph.objects(test_uri, MF.requires))
    has_star = JELLYT.requirementRdfStar in reqs
    has_gen = JELLYT.requirementGeneralizedRdf in reqs

    if has_star and has_gen:
        return "rdf_star_generalized"
    if has_star:
        return "rdf_star"
    if has_gen:
        return "generalized"
    return "rdf11"



from typing import Any, Final, cast

import pytest

from pyjelly import jelly
from pyjelly.errors import JellyAssertionError
from pyjelly.integrations.generic.serialize import guess_options, guess_stream
from pyjelly.options import StreamParameters
from pyjelly.serialize.streams import QuadStream, SerializerOptions, TripleStream


class _Sink:
    def __init__(self, *, is_triples_sink: bool) -> None:
        self.is_triples_sink: Final[bool] = is_triples_sink


@pytest.mark.parametrize(
    ("sink", "expected_logical"),
    [
        (_Sink(is_triples_sink=True), jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (_Sink(is_triples_sink=False), jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
    ],
    ids=["triples-sink", "quads-sink"],
)
def test_defaults_generic(
    sink: Any,
    expected_logical: int,
) -> None:
    opts = guess_options(cast(Any, sink))
    assert opts.logical_type == expected_logical
    assert opts.params.rdf_star is True
    assert opts.params.generalized_statements is True
    assert opts.params.namespace_declarations is False


@pytest.mark.parametrize(
    ("sink", "logical", "expected_physical"),
    [
        (
            _Sink(is_triples_sink=True),
            jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            TripleStream,
        ),
        (
            _Sink(is_triples_sink=False),
            jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            QuadStream,
        ),
    ],
    ids=["triples-sink+triples", "quads-sink+quads"],
)
def test_override_generic_compatible(
    sink: Any,
    logical: int,
    expected_physical: type,
) -> None:
    user_opts = SerializerOptions(
        logical_type=cast(jelly.LogicalStreamType, logical),
        params=StreamParameters(
            rdf_star=False, generalized_statements=False, namespace_declarations=False
        ),
    )
    stream = guess_stream(user_opts, cast(Any, sink))
    assert isinstance(stream, expected_physical)
    assert stream.options.logical_type == logical
    assert stream.options.params.rdf_star is False
    assert stream.options.params.generalized_statements is False


@pytest.mark.parametrize(
    ("sink", "logical"),
    [
        (_Sink(is_triples_sink=True), jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (_Sink(is_triples_sink=False), jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
    ],
    ids=["triples-sink+quads-error", "quads-sink+triples-error"],
)
def test_override_generic_incompatible(
    sink: Any,
    logical: int,
) -> None:
    user_opts = SerializerOptions(
        logical_type=cast(jelly.LogicalStreamType, logical),
        params=StreamParameters(
            rdf_star=False, generalized_statements=False, namespace_declarations=False
        ),
    )
    with pytest.raises(
        JellyAssertionError,
        match="is not compatible with",
    ):
        guess_stream(user_opts, cast(Any, sink))



from __future__ import annotations

import io

from pyjelly import jelly
from pyjelly.integrations.generic.generic_sink import (
    GenericStatementSink,
    Prefix,
)
from pyjelly.integrations.generic.parse import parse_jelly_flat
from pyjelly.integrations.generic.serialize import flat_stream_to_file
from pyjelly.options import LookupPreset
from pyjelly.parse.ioutils import get_options_and_frames
from pyjelly.serialize.flows import FlatQuadsFrameFlow, FlatTriplesFrameFlow
from pyjelly.serialize.streams import SerializerOptions


class GenericSerDes:
    def __init__(self) -> None:
        self.name = "generic"

    def read_quads(self, in_bytes: bytes) -> GenericStatementSink:
        bio = io.BytesIO(in_bytes)
        options, frames = get_options_and_frames(bio)
        sink = GenericStatementSink()
        for item in parse_jelly_flat(inp=bio, frames=frames, options=options):
            if isinstance(item, Prefix):
                sink.bind(item.prefix, item.iri)
            else:
                sink.add(item)
        return sink

    def write_quads(self, in_graph: GenericStatementSink) -> bytes:
        out = io.BytesIO()
        options = SerializerOptions(logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS)
        flat_stream_to_file(in_graph.store, out, options=options)
        return out.getvalue()

    def write_quads_jelly(
        self, in_graph: GenericStatementSink, preset: LookupPreset, frame_size: int
    ) -> bytes:
        out = io.BytesIO()
        options = SerializerOptions(
            flow=FlatQuadsFrameFlow(frame_size=frame_size),
            logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            lookup_preset=preset,
        )
        flat_stream_to_file(in_graph.store, out, options=options)
        return out.getvalue()

    def read_triples(self, in_bytes: bytes) -> GenericStatementSink:
        bio = io.BytesIO(in_bytes)
        options, frames = get_options_and_frames(bio)
        sink = GenericStatementSink()
        for item in parse_jelly_flat(inp=bio, frames=frames, options=options):
            if isinstance(item, Prefix):
                sink.bind(item.prefix, item.iri)
            else:
                sink.add(item)
        return sink

    def write_triples(self, in_graph: GenericStatementSink) -> bytes:
        out = io.BytesIO()
        options = SerializerOptions(logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES)
        flat_stream_to_file(in_graph.store, out, options=options)
        return out.getvalue()

    def write_triples_jelly(
        self, in_graph: GenericStatementSink, preset: LookupPreset, frame_size: int
    ) -> bytes:
        out = io.BytesIO()
        options = SerializerOptions(
            flow=FlatTriplesFrameFlow(frame_size=frame_size),
            logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            lookup_preset=preset,
        )
        flat_stream_to_file(in_graph.store, out, options=options)
        return out.getvalue()



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from unittest.mock import patch

import pytest
from rdflib import Dataset, Graph, Node, URIRef
from rdflib import Literal as RdfLiteral
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.namespace import RDF
from rdflib.plugins.serializers.nt import _quoteLiteral

from pyjelly.integrations.generic.parse import (
    parse_jelly_grouped as generic_parse_jelly_grouped,
)
from pyjelly.integrations.rdflib.parse import parse_jelly_grouped
from tests.conformance_tests.test_rdf._common import (
    JELLYT,
    MF,
    categorize_by_requires,
)
from tests.meta import RDF_FROM_JELLY_MANIFEST, TEST_OUTPUTS_DIR
from tests.utils.generic_sink_test_serializer import GenericSinkSerializer
from tests.utils.ordered_memory import OrderedMemory
from tests.utils.rdf_test_cases import jelly_validate, needs_jelly_cli

# Manifest path for from_jelly tests. This is where all test cases are defined.
FROM_JELLY_MANIFEST = RDF_FROM_JELLY_MANIFEST


@dataclass
class FromJellyTestCase:
    uri: str
    name: str
    action_path: Path
    result_paths: list[Path] | None
    test_type: str
    category: str
    id: str = field(init=False)

    def __post_init__(self) -> None:
        # Custom id for pytest output to easily see type/category/case
        self.id = f"{self.test_type}-{self.category}-{self.action_path.parent.name}"


# Load all test cases from the manifest file
# Important: This function translates the RDF manifest
# into Python test cases.
# If manifests change their structure (e.g. new predicates), update logic here.
def load_from_jelly_manifest_cases(manifest_path: Path) -> list[FromJellyTestCase]:
    if not manifest_path.exists():
        return []
    graph = Graph()
    graph.parse(manifest_path, format="turtle")
    manifest_dir = manifest_path.parent
    base_uri_from_manifest = "https://w3id.org/jelly/dev/tests/rdf/from_jelly/"
    test_cases = []
    test_type_map = {
        JELLYT.TestPositive: "positive",
        JELLYT.TestNegative: "negative",
    }
    for test_class, test_type_str in test_type_map.items():
        for test_uri in graph.subjects(RDF.type, test_class):
            if not isinstance(test_uri, URIRef):
                continue
            # Map MF.action to the actual input file path
            action_uri = graph.value(test_uri, MF.action)
            action_rel_path = str(action_uri).replace(base_uri_from_manifest, "")
            action_path = manifest_dir / action_rel_path

            # MF.result can be a single file or an RDF list of files
            result_paths = None
            result_node = graph.value(test_uri, MF.result)
            if result_node:
                if (result_node, RDF.first, None) in graph:
                    # Handle list of result files
                    result_uris = graph.items(result_node)
                    result_paths = [
                        manifest_dir / str(uri).replace(base_uri_from_manifest, "")
                        for uri in result_uris
                    ]
                else:
                    # Single result file
                    result_rel_path = str(result_node).replace(
                        base_uri_from_manifest, ""
                    )
                    result_paths = [manifest_dir / result_rel_path]

            # Each test case knows its category (rdf11/generalized/etc.)
            # from categorize_by_requires
            test_cases.append(
                FromJellyTestCase(
                    uri=str(test_uri),
                    name=str(graph.value(test_uri, MF.name) or ""),
                    action_path=action_path,
                    result_paths=result_paths,
                    test_type=test_type_str,
                    category=categorize_by_requires(graph, test_uri),
                )
            )
    return test_cases


# Patch rdflib nquads serializer to avoid writing default graph id
# Without this patch rdflib includes the default graph id explicitly,
# which breaks comparisons.
def _new_nq_row(triple: tuple[Node, Node, Node], context: Graph) -> str:
    template = "%s " * (3 + (context != DATASET_DEFAULT_GRAPH_ID)) + ".\n"
    args = (
        triple[0].n3(),
        triple[1].n3(),
        _quoteLiteral(triple[2])
        if isinstance(triple[2], RdfLiteral)
        else triple[2].n3(),
        *((context.n3(),) if context != DATASET_DEFAULT_GRAPH_ID else ()),
    )
    return template % args


workaround_rdflib_serializes_default_graph_id = patch(
    "rdflib.plugins.serializers.nquads._nq_row",
    new=_new_nq_row,
)
workaround_rdflib_serializes_default_graph_id.start()

# Collect and categorize test cases for pytest parameterization
ALL_CASES = load_from_jelly_manifest_cases(FROM_JELLY_MANIFEST)

RDF11_POSITIVE_CASES = [
    pytest.param(case, id=case.id)
    for case in ALL_CASES
    if case.test_type == "positive" and case.category == "rdf11"
]
ALL_POSITIVE_CASES = [
    pytest.param(case, id=case.id) for case in ALL_CASES if case.test_type == "positive"
]
RDF11_NEGATIVE_CASES = [
    pytest.param(case, id=case.id)
    for case in ALL_CASES
    if case.test_type == "negative" and case.category == "rdf11"
]
ALL_NEGATIVE_CASES = [
    pytest.param(case, id=case.id) for case in ALL_CASES if case.test_type == "negative"
]


# RDF11 positive tests using rdflib parser.
# Ensures rdflib correctly parses RDF 1.1 cases.
@needs_jelly_cli
@pytest.mark.parametrize("case", RDF11_POSITIVE_CASES)
def test_rdflib_parses_rdf11_positive(case: FromJellyTestCase) -> None:
    test_id = case.action_path.parent.name
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True, parents=True)
    with case.action_path.open("rb") as input_file:
        for frame_no, graph in enumerate(
            parse_jelly_grouped(
                input_file,
                graph_factory=lambda: Graph(store=OrderedMemory()),
                dataset_factory=lambda: Dataset(store=OrderedMemory()),
            )
        ):
            output_format = "nquads" if isinstance(graph, Dataset) else "ntriples"
            file_extension = ".nq" if isinstance(graph, Dataset) else ".nt"
            output_filename = output_dir / f"out_{frame_no:03}{file_extension}"
            graph.serialize(
                destination=output_filename, encoding="utf-8", format=output_format
            )
            # Validate against reference result
            jelly_validate(
                case.action_path,
                "--compare-ordered",
                "--compare-frame-indices",
                frame_no,
                "--compare-to-rdf-file",
                output_filename,
                hint=f"Test ID: {test_id}, output file: {output_filename}",
            )


# Generic integration positive tests.
# Ensure our generic integration behaves the same as reference.
@needs_jelly_cli
@pytest.mark.parametrize("case", ALL_POSITIVE_CASES)
def test_generic_parses_positive(case: FromJellyTestCase) -> None:
    test_id = case.action_path.parent.name
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True, parents=True)
    with case.action_path.open("rb") as input_file:
        for frame_no, graph in enumerate(generic_parse_jelly_grouped(input_file)):
            file_extension = case.result_paths[0].suffix if case.result_paths else ".nt"
            output_filename = output_dir / f"out_{frame_no:03}{file_extension}"
            serializer = GenericSinkSerializer(graph)
            serializer.serialize(output_filename=output_filename, encoding="utf-8")
            # Validate against reference result
            jelly_validate(
                case.action_path,
                "--compare-ordered",
                "--compare-frame-indices",
                frame_no,
                "--compare-to-rdf-file",
                output_filename,
                hint=f"Test ID: {test_id}, output file: {output_filename}",
            )


# RDF11 negative tests using rdflib parser. These ensure parser fails where it should.
@needs_jelly_cli
@pytest.mark.parametrize("case", RDF11_NEGATIVE_CASES)
def test_rdflib_parsing_fails_rdf11_negative(case: FromJellyTestCase) -> None:
    test_id = case.action_path.parent.name
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True, parents=True)
    dataset = Dataset(store=OrderedMemory())
    # We expect an error for negative cases
    with pytest.raises(Exception, match=".*"):
        dataset.parse(location=str(case.action_path), format="jelly")


# Generic integration negative tests. Same idea: must fail consistently.
@needs_jelly_cli
@pytest.mark.parametrize("case", ALL_NEGATIVE_CASES)
def test_generic_parsing_fails_negative(case: FromJellyTestCase) -> None:
    test_id = case.action_path.parent.name
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True, parents=True)
    with (
        pytest.raises(Exception, match=".*"),
        case.action_path.open("rb") as input_file,
    ):
        list(generic_parse_jelly_grouped(input_file))



from __future__ import annotations

from pyjelly import jelly
from pyjelly.parse.decode import options_from_frame


def _frame_with_options(
    *, version: int, set_nd: bool | None
) -> tuple[jelly.RdfStreamFrame, jelly.RdfStreamOptions]:
    frame = jelly.RdfStreamFrame()

    opts = jelly.RdfStreamOptions()
    opts.version = version
    opts.stream_name = "t"
    opts.physical_type = jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    opts.logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
    opts.max_name_table_size = 128
    opts.max_prefix_table_size = 32
    opts.max_datatype_table_size = 32

    if set_nd is not None and hasattr(opts, "namespace_declarations"):
        opts.namespace_declarations = set_nd

    row = frame.rows.add()
    row.options.CopyFrom(opts)

    return frame, opts


def test_nd_inferred_true_when_version_is_v2_and_field_missing() -> None:
    frame, _ = _frame_with_options(version=2, set_nd=None)
    parser_opts = options_from_frame(frame, delimited=True)
    assert parser_opts.params.namespace_declarations is True
    assert parser_opts.params.version == 2


def test_nd_false_when_version_is_v1_and_field_missing() -> None:
    frame, _ = _frame_with_options(version=1, set_nd=None)
    parser_opts = options_from_frame(frame, delimited=True)
    assert parser_opts.params.namespace_declarations is False
    assert parser_opts.params.version == 1



from rdflib import Graph, Namespace, URIRef, Literal
from pyjelly.integrations.rdflib.serialize import SerializerOptions, StreamParameters

# Build a tiny graph and bind a prefix
g = Graph()
EX = Namespace("http://example.org/")
g.namespace_manager.bind("ex", EX)
g.add((EX.alice, URIRef("http://xmlns.com/foaf/0.1/name"), Literal("Alice")))

print("IN  namespaces:", dict(g.namespaces()))

# Enable namespace declarations in Jelly output
options = SerializerOptions(params=StreamParameters(namespace_declarations=True))

# Serialize with options
g.serialize("sample_test.jelly", format="jelly", options=options)

# Parse back and check namespaces
g_new = Graph()
g_new.parse("sample_test.jelly", format="jelly")
print("OUT namespaces:", dict(g_new.namespaces()))



from __future__ import annotations

from pyjelly import jelly
from pyjelly.parse.decode import Adapter, Decoder, options_from_frame


def _frame_with_options(
    *, version: int, set_nd: bool | None
) -> tuple[jelly.RdfStreamFrame, jelly.RdfStreamOptions]:
    frame = jelly.RdfStreamFrame()

    opts = jelly.RdfStreamOptions()
    opts.version = version
    opts.stream_name = "t"
    opts.physical_type = jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    opts.logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
    opts.max_name_table_size = 128
    opts.max_prefix_table_size = 32
    opts.max_datatype_table_size = 32

    if set_nd is not None and hasattr(opts, "namespace_declarations"):
        opts.namespace_declarations = set_nd

    row = frame.rows.add()
    row.options.CopyFrom(opts)
    return frame, opts


class _DummyAdapter(Adapter):
    def iri(self, iri: str) -> str:
        return iri

    def default_graph(self) -> None:
        return None

    def bnode(self, bnode: str) -> str:
        return bnode

    def literal(
        self, lex: str, language: str | None = None, datatype: str | None = None
    ) -> tuple[str, str | None, str | None]:
        return (lex, language, datatype)


def test_streamtypes_compat_with_v2_and_inferred_nd() -> None:
    frame, raw_opts = _frame_with_options(version=2, set_nd=None)
    parser_opts = options_from_frame(frame, delimited=True)

    dec = Decoder(_DummyAdapter(parser_opts))
    dec.validate_stream_options(raw_opts)


def test_streamtypes_compat_with_v1_and_no_nd() -> None:
    frame, raw_opts = _frame_with_options(version=1, set_nd=None)
    parser_opts = options_from_frame(frame, delimited=True)

    dec = Decoder(_DummyAdapter(parser_opts))
    dec.validate_stream_options(raw_opts)



import pytest
from rdflib import URIRef

from pyjelly import jelly
from pyjelly.integrations.rdflib.serialize import RDFLibTermEncoder
from pyjelly.options import LookupPreset
from pyjelly.serialize.encode import Slot


def test_graph_not_implemented() -> None:
    enc = RDFLibTermEncoder(lookup_preset=LookupPreset())
    with pytest.raises(NotImplementedError, match="unsupported term type"):
        enc.encode_graph(
            (URIRef("http://ex/s1"), URIRef("http://ex/p1"), URIRef("http://ex/o1")),
            jelly.RdfQuad(),
        )


def test_spo_not_implemented() -> None:
    enc = RDFLibTermEncoder(lookup_preset=LookupPreset())
    with pytest.raises(NotImplementedError, match="unsupported term type"):
        enc.encode_spo(
            (URIRef("http://ex/s1"), URIRef("http://ex/p1"), URIRef("http://ex/o1")),
            Slot.subject,
            jelly.RdfQuad(),
        )



from __future__ import annotations

import io
from contextlib import nullcontext
from dataclasses import dataclass
from typing import Any, Callable
from unittest.mock import patch

import pytest
from rdflib import Dataset, Graph
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.integrations.rdflib.parse import parse_jelly_flat, parse_jelly_grouped
from pyjelly.parse.ioutils import get_options_and_frames
from pyjelly.serialize.streams import GraphStream, QuadStream, SerializerOptions, Stream


def test_flat_triples() -> None:
    g_out = Graph()
    g_out.parse(source="tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")

    out = g_out.serialize(encoding="jelly", format="jelly")
    triples_out = set(g_out)
    assert len(triples_out) > 0

    g_in = Graph()
    g_in.parse(out, format="jelly")

    triples_in = set(g_in)

    assert len(triples_out) == len(triples_in)
    assert triples_in == triples_out

    options, _ = get_options_and_frames(io.BytesIO(out))
    assert options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    assert options.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES

    ds_in = Dataset()
    ds_in.parse(out, format="jelly")

    quads_in = set(ds_in)
    assert quads_in == {(*triple, DATASET_DEFAULT_GRAPH_ID) for triple in triples_in}


@pytest.mark.parametrize("stream_class", [QuadStream, GraphStream])
def test_flat_quads(stream_class: type[Stream]) -> None:
    ds_out = Dataset()
    ds_out.parse(source="tests/e2e_test_cases/quads_rdf_1_1/weather-quads.nq")

    stream = stream_class.for_rdflib()

    out = ds_out.serialize(encoding="jelly", format="jelly", stream=stream)
    quads_out = set(ds_out)
    assert len(quads_out) > 0

    options, _ = get_options_and_frames(io.BytesIO(out))
    assert options.stream_types.physical_type == stream_class.physical_type
    assert options.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS

    ds_in = Dataset()
    ds_in.parse(out, format="jelly")

    quads_in = set(ds_in)
    assert quads_in == quads_out


@pytest.mark.skip
def test_graphs() -> None:
    # TODO(Nastya): rewrite or remove
    options = SerializerOptions(logical_type=jelly.LOGICAL_STREAM_TYPE_GRAPHS)

    ds_out = Dataset()
    g1_out = Graph(identifier="foaf")
    g1_out.parse(source="tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")
    g2_out = Graph(identifier="test")
    g2_out.parse(source="tests/e2e_test_cases/triples_rdf_1_1/p2_ontology.nt")
    ds_out.add_graph(g1_out)
    ds_out.add_graph(g2_out)

    out = ds_out.serialize(options=options, encoding="jelly", format="jelly")

    options_out, _ = get_options_and_frames(io.BytesIO(out))
    assert options_out.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    assert options_out.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_GRAPHS

    graphs_out = sorted(ds_out.graphs(), key=len)

    ds_in = Dataset()
    ds_in.parse(out, format="jelly")

    graphs_in = sorted(ds_in.graphs(), key=len)

    for g_out, g_in in zip(graphs_out, graphs_in):
        assert len(g_out) == len(g_in)
        assert set(g_out) == set(g_in)


def _make_flat_triples_bytes() -> bytes:
    g = Graph()
    g.parse("tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")
    return g.serialize(encoding="jelly", format="jelly")


def _make_flat_quads_bytes() -> bytes:
    ds = Dataset()
    ds.parse("tests/e2e_test_cases/quads_rdf_1_1/weather-quads.nq")
    return ds.serialize(
        encoding="jelly",
        format="jelly",
        stream=QuadStream.for_rdflib(),
    )


def _make_physical_graphs_bytes() -> bytes:
    ds = Dataset()
    g1 = Graph(identifier="g1")
    g1.parse("tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")
    g2 = Graph(identifier="g2")
    g2.parse("tests/e2e_test_cases/triples_rdf_1_1/p2_ontology.nt")
    ds.add_graph(g1)
    ds.add_graph(g2)

    return ds.serialize(
        encoding="jelly", format="jelly", stream=GraphStream.for_rdflib()
    )


def _make_grouped_graphs_bytes() -> bytes:
    ds = Dataset()
    g1 = Graph(identifier="g1")
    g1.parse("tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")
    g2 = Graph(identifier="g2")
    g2.parse("tests/e2e_test_cases/triples_rdf_1_1/p2_ontology.nt")
    ds.add_graph(g1)
    ds.add_graph(g2)
    return ds.serialize(
        options=SerializerOptions(logical_type=jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        encoding="jelly",
        format="jelly",
    )


@dataclass(frozen=True)
class Case:
    make_bytes: Callable[[], bytes]
    expected_physical: int
    expected_logical: int
    parser: Callable[..., Any]
    strict: bool
    should_raise: bool
    msg: str | None


@pytest.mark.parametrize(
    "case",
    [
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=parse_jelly_grouped,
            strict=True,
            should_raise=True,
            msg="expected GROUPED",
        ),
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=parse_jelly_grouped,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=parse_jelly_flat,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=parse_jelly_flat,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_grouped_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_GRAPHS,
            parser=parse_jelly_flat,
            strict=True,
            should_raise=True,
            msg="expected FLAT",
        ),
        Case(
            make_bytes=_make_grouped_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_GRAPHS,
            parser=parse_jelly_grouped,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=parse_jelly_grouped,
            strict=True,
            should_raise=True,
            msg="expected GROUPED",
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=parse_jelly_grouped,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=parse_jelly_flat,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=parse_jelly_flat,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=parse_jelly_grouped,
            strict=True,
            should_raise=True,
            msg="expected GROUPED",
        ),
    ],
)
def test_rdflib_logical_matrix(case: Case) -> None:
    data = case.make_bytes()
    opts, _ = get_options_and_frames(io.BytesIO(data))
    assert opts.stream_types.physical_type == case.expected_physical
    assert opts.stream_types.logical_type == case.expected_logical

    ctx = (
        pytest.raises(JellyConformanceError, match=case.msg)
        if case.should_raise
        else nullcontext()
    )
    with ctx:
        list(case.parser(io.BytesIO(data), logical_type_strict=case.strict))


def test_rdflib_flat_strict_requires_stream_types() -> None:
    class Opt:
        stream_types = None

    dummy = b"x"
    frames: list[object] = []

    with (
        patch(
            "pyjelly.integrations.rdflib.parse.get_options_and_frames",
            return_value=(Opt(), frames),
        ),
        pytest.raises(JellyConformanceError, match="requires options.stream_types"),
    ):
        list(parse_jelly_flat(io.BytesIO(dummy), logical_type_strict=True))


def test_rdflib_grouped_strict_unspecified_raises() -> None:
    class ST:
        flat = False
        logical_type = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
        physical_type = jelly.PHYSICAL_STREAM_TYPE_TRIPLES

    dummy = b"x"
    options = type("Opt", (), {"stream_types": ST()})()
    frames: list[object] = []

    with (
        patch(
            "pyjelly.integrations.rdflib.parse.get_options_and_frames",
            return_value=(options, frames),
        ),
        pytest.raises(JellyConformanceError, match="expected GROUPED"),
    ):
        list(parse_jelly_grouped(io.BytesIO(dummy), logical_type_strict=True))


def test_rdflib_flat_unsupported_physical_raises() -> None:
    class ST:
        flat = True
        logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
        physical_type = jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED

    dummy = b"x"
    options = type("Opt", (), {"stream_types": ST()})()
    frames: list[object] = []

    with (
        patch(
            "pyjelly.integrations.rdflib.parse.get_options_and_frames",
            return_value=(options, frames),
        ),
        pytest.raises(NotImplementedError, match="not supported"),
    ):
        list(parse_jelly_flat(io.BytesIO(dummy), logical_type_strict=False))



# tests/unit/test_generic_serialize.py
from __future__ import annotations

from collections.abc import Generator

import pytest

from pyjelly import jelly
from pyjelly.integrations.generic.generic_sink import (
    IRI,
    GenericStatementSink,
    Literal,
    Quad,
    Triple,
)
from pyjelly.integrations.generic.serialize import (
    GenericSinkTermEncoder,
    flat_stream_to_frames,
    graphs_stream_frames,
    grouped_stream_to_frames,
    guess_options,
    guess_stream,
    quads_stream_frames,
    split_to_graphs,
    stream_frames,
    triples_stream_frames,
)
from pyjelly.options import LookupPreset, StreamParameters
from pyjelly.serialize.encode import Slot
from pyjelly.serialize.flows import (
    DatasetsFrameFlow,
    FlatQuadsFrameFlow,
)
from pyjelly.serialize.streams import (
    GraphStream,
    QuadStream,
    SerializerOptions,
    TripleStream,
)


def test_stream_frames_typeerror() -> None:
    with pytest.raises(TypeError):
        list(stream_frames(object(), GenericStatementSink()))  # type: ignore[arg-type]


def test_split_to_graphs() -> None:
    g1 = IRI("http://g1")
    g2 = IRI("http://g2")

    def gen() -> Generator[Quad]:
        yield Quad(IRI("http://s1"), IRI("http://p1"), IRI("http://o1"), g1)
        yield Quad(IRI("http://s2"), IRI("http://p2"), IRI("http://o2"), g1)
        yield Quad(IRI("http://s3"), IRI("http://p3"), IRI("http://o3"), g2)

    graphs = list(split_to_graphs(gen()))
    assert len(graphs) == 2
    assert graphs[0].identifier == g1
    assert graphs[1].identifier == g2


def test_guess_options_and_stream() -> None:
    def _triple_sink() -> GenericStatementSink:
        s = GenericStatementSink()
        s.bind("ex", IRI("http://example.com/"))
        s.add(Triple(IRI("http://ex/s1"), IRI("http://ex/p1"), IRI("http://ex/o1")))
        s.add(
            Triple(IRI("http://ex/s2"), IRI("http://ex/p2"), Literal("example_value"))
        )
        return s

    def _quad_sink() -> GenericStatementSink:
        s = GenericStatementSink()
        s.bind("ex", IRI("http://example.com/"))
        g1 = IRI("http://ex/g1")
        g2 = IRI("http://ex/g2")
        s.add(
            Quad(IRI("http://ex/sg1"), IRI("http://ex/pg1"), IRI("http://ex/og1"), g1)
        )
        s.add(
            Quad(IRI("http://ex/sg2"), IRI("http://ex/pg2"), IRI("http://ex/og2"), g2)
        )
        return s

    t_sink = _triple_sink()
    q_sink = _quad_sink()
    t_opts = guess_options(t_sink)
    q_opts = guess_options(q_sink)
    assert t_opts.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
    assert q_opts.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS
    assert isinstance(guess_stream(t_opts, t_sink), TripleStream)
    assert isinstance(guess_stream(q_opts, q_sink), QuadStream)


def test_flat_stream_to_frames_empty_generator() -> None:
    def empty_statements() -> Generator[Triple | Quad]:
        if False:
            yield
        return

    frames = list(flat_stream_to_frames(empty_statements()))
    assert frames == []


def test_flat_stream_guesses_options() -> None:
    def gen() -> Generator[Triple]:
        yield Triple(IRI("http://s"), IRI("http://p"), Literal("http://o"))

    frames = list(flat_stream_to_frames(gen()))
    assert frames
    assert isinstance(frames[-1], jelly.RdfStreamFrame)


def test_graphs_stream_frames_from_generator() -> None:
    options = SerializerOptions()
    stream = GraphStream(
        encoder=GenericSinkTermEncoder(),
        options=options,
    )

    def gen() -> Generator[Quad]:
        yield Quad(
            IRI("http://s1"), IRI("http://p1"), IRI("http://o1"), IRI("http://g1")
        )

    out = list(graphs_stream_frames(stream, gen()))
    assert out
    assert isinstance(out[-1], jelly.RdfStreamFrame)


@pytest.mark.parametrize("with_namespace", [True, False])
def test_triples_stream_frames_parametrized(*, with_namespace: bool) -> None:
    options = SerializerOptions(
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        params=StreamParameters(namespace_declarations=with_namespace),
    )
    stream = TripleStream(
        encoder=GenericSinkTermEncoder(lookup_preset=LookupPreset()),
        options=options,
    )

    sink = GenericStatementSink()
    if with_namespace:
        sink.bind("ex", IRI("http://example.com/"))
    sink.add(Triple(IRI("http://ex/s"), IRI("http://ex/p"), Literal("http://ex/o")))

    frames = list(triples_stream_frames(stream, sink))
    assert frames
    assert isinstance(frames[-1], jelly.RdfStreamFrame)


@pytest.mark.parametrize("with_namespace", [True, False])
def test_quads_stream_frames_parametrized(*, with_namespace: bool) -> None:
    options = SerializerOptions(
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
        params=StreamParameters(namespace_declarations=with_namespace),
    )
    stream = QuadStream(
        encoder=GenericSinkTermEncoder(lookup_preset=LookupPreset()),
        options=options,
    )

    sink = GenericStatementSink()
    if with_namespace:
        sink.bind("ex", IRI("http://example.com/"))
    sink.add(
        Quad(
            IRI("http://ex/s1"),
            IRI("http://ex/p1"),
            IRI("http://ex/o1"),
            IRI("http://ex/g1"),
        )
    )
    sink.add(
        Quad(
            IRI("http://ex/s2"),
            IRI("http://ex/p2"),
            IRI("http://ex/o2"),
            IRI("http://ex/g2"),
        )
    )

    out = list(quads_stream_frames(stream, sink))
    assert out
    assert isinstance(out[-1], jelly.RdfStreamFrame)


@pytest.mark.parametrize("with_namespace", [True, False])
def test_graphs_stream_frames_parametrized(*, with_namespace: bool) -> None:
    options = SerializerOptions(
        params=StreamParameters(namespace_declarations=with_namespace),
    )
    stream = GraphStream(
        encoder=GenericSinkTermEncoder(lookup_preset=LookupPreset()),
        options=options,
    )

    sink = GenericStatementSink()
    if with_namespace:
        sink.bind("ex", IRI("http://example.com/"))
    sink.add(
        Quad(
            IRI("http://ex/s1"),
            IRI("http://ex/p1"),
            IRI("http://ex/o1"),
            IRI("http://ex/g1"),
        )
    )
    sink.add(
        Quad(
            IRI("http://ex/s2"),
            IRI("http://ex/p2"),
            IRI("http://ex/o2"),
            IRI("http://ex/g2"),
        )
    )

    out = list(graphs_stream_frames(stream, sink))
    assert out
    assert isinstance(out[-1], jelly.RdfStreamFrame)


def test_encoder_unsupported_raises() -> None:
    enc = GenericSinkTermEncoder(lookup_preset=LookupPreset())
    with pytest.raises(NotImplementedError, match="unsupported term type"):
        enc.encode_spo(object(), Slot.subject, jelly.RdfTriple())


def test_graphs_stream_frames_emit_dataset() -> None:
    opts = SerializerOptions(
        flow=DatasetsFrameFlow(),
        logical_type=jelly.LOGICAL_STREAM_TYPE_DATASETS,
    )
    stream = GraphStream(
        encoder=GenericSinkTermEncoder(lookup_preset=LookupPreset()),
        options=opts,
    )

    sink = GenericStatementSink()
    sink.add(
        Quad(IRI("http://s1"), IRI("http://p1"), IRI("http://o1"), IRI("http://g1"))
    )
    sink.add(
        Quad(IRI("http://s2"), IRI("http://p2"), IRI("http://o2"), IRI("http://g2"))
    )
    frames = list(graphs_stream_frames(stream, sink))
    assert frames
    assert isinstance(frames[-1], jelly.RdfStreamFrame)


def test_graphs_stream_frames_emit_flat() -> None:
    sink = GenericStatementSink()
    sink.add(
        Quad(IRI("http://s1"), IRI("http://p1"), IRI("http://o1"), IRI("http://g1"))
    )
    sink.add(
        Quad(IRI("http://s2"), IRI("http://p2"), IRI("http://o2"), IRI("http://g2"))
    )

    opts = SerializerOptions(
        flow=FlatQuadsFrameFlow(),
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
    )
    stream = GraphStream(
        encoder=GenericSinkTermEncoder(lookup_preset=LookupPreset()),
        options=opts,
    )
    frames = list(graphs_stream_frames(stream, sink))
    assert frames
    assert isinstance(frames[-1], jelly.RdfStreamFrame)


def test_grouped_stream_to_frames_init_stream_guess_options() -> None:
    s1 = GenericStatementSink()
    s1.add(Triple(IRI("http://ex/s1"), IRI("http://ex/p1"), Literal("http://ex/o1")))
    s2 = GenericStatementSink()
    s2.add(Triple(IRI("http://ex/s2"), IRI("http://ex/p2"), Literal("http://ex/o2")))

    def gen() -> Generator[GenericStatementSink, None, None]:
        yield s1
        yield s2

    frames = list(grouped_stream_to_frames(gen(), options=None))
    expected = 2
    assert len(frames) == expected


def test_graph_not_implemented() -> None:
    enc = GenericSinkTermEncoder(lookup_preset=LookupPreset())
    with pytest.raises(NotImplementedError, match="unsupported term type"):
        enc.encode_graph(
            Triple(IRI("http://ex/s1"), IRI("http://ex/p1"), IRI("http://ex/o1")),
            jelly.RdfQuad(),
        )



from __future__ import annotations

from collections.abc import Generator, Iterable
from dataclasses import dataclass, field
from typing import ClassVar

from pyjelly import jelly
from pyjelly.options import LookupPreset, StreamParameters, StreamTypes
from pyjelly.serialize.encode import (
    Slot,
    TermEncoder,
    encode_namespace_declaration,
    encode_options,
    encode_quad,
    encode_triple,
)
from pyjelly.serialize.flows import (
    DEFAULT_FRAME_SIZE,
    BoundedFrameFlow,
    FlatQuadsFrameFlow,
    FlatTriplesFrameFlow,
    FrameFlow,
    ManualFrameFlow,
    flow_for_type,
)


@dataclass
class SerializerOptions:
    flow: FrameFlow | None = None
    frame_size: int = DEFAULT_FRAME_SIZE
    logical_type: jelly.LogicalStreamType = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
    params: StreamParameters = field(default_factory=StreamParameters)
    lookup_preset: LookupPreset = field(default_factory=LookupPreset)


class Stream:
    physical_type: ClassVar[jelly.PhysicalStreamType]
    default_delimited_flow_class: ClassVar[type[BoundedFrameFlow]]
    repeated_terms: list[object | None]

    def __init__(
        self,
        *,
        encoder: TermEncoder,
        options: SerializerOptions | None = None,
    ) -> None:
        self.encoder = encoder
        if options is None:
            options = SerializerOptions()
        self.options = options
        flow = options.flow
        if flow is None:
            flow = self.infer_flow()
        self.flow = flow
        self.repeated_terms = [None] * len(Slot)
        self.enrolled = False
        self.stream_types = StreamTypes(
            physical_type=self.physical_type,
            logical_type=self.flow.logical_type,
        )

    def infer_flow(self) -> FrameFlow:
        """
        Return flow based on the stream options provided.

        Returns:
            FrameFlow: initialised FrameFlow object.

        """
        flow: FrameFlow
        if self.options.params.delimited:
            if self.options.logical_type != jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED:
                flow_class = flow_for_type(self.options.logical_type)
            else:
                flow_class = self.default_delimited_flow_class

            if self.options.logical_type in (
                jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
                jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            ):
                flow = flow_class(
                    logical_type=self.options.logical_type,
                    frame_size=self.options.frame_size,
                )
            else:
                flow = flow_class(logical_type=self.options.logical_type)
        else:
            flow = ManualFrameFlow(logical_type=self.options.logical_type)
        return flow

    def enroll(self) -> None:
        """Initialize start of the stream."""
        if not self.enrolled:
            self.stream_options()
            self.enrolled = True

    def stream_options(self) -> None:
        """Encode and append stream options row to the current flow."""
        self.flow.append(
            encode_options(
                stream_types=self.stream_types,
                params=self.options.params,
                lookup_preset=self.options.lookup_preset,
            )
        )

    def namespace_declaration(self, name: str, iri: str) -> None:
        """
        Add namespace declaration to jelly stream.

        Args:
            name (str): namespace prefix label
            iri (str): namespace iri

        """
        rows = encode_namespace_declaration(
            name=name,
            value=iri,
            term_encoder=self.encoder,
        )
        self.flow.extend(rows)

    @classmethod
    def for_rdflib(cls, options: SerializerOptions | None = None) -> Stream:
        """
        Initialize stream with RDFLib encoder.

        Args:
            options (SerializerOptions | None, optional): Stream options.
                Defaults to None.

        Raises:
            TypeError: if Stream is passed, and not a Stream for specific physical type.

        Returns:
            Stream: initialized stream with RDFLib encoder.

        """
        if cls is Stream:
            msg = "Stream is an abstract base class, use a subclass instead"
            raise TypeError(msg)
        from pyjelly.integrations.rdflib.serialize import RDFLibTermEncoder

        lookup_preset: LookupPreset | None = None
        if options is not None:
            lookup_preset = options.lookup_preset
        return cls(
            encoder=RDFLibTermEncoder(lookup_preset=lookup_preset),
            options=options,
        )


def stream_for_type(physical_type: jelly.PhysicalStreamType) -> type[Stream]:
    """
    Give a Stream based on physical type specified.

    Args:
        physical_type (jelly.PhysicalStreamType): jelly stream physical type.

    Raises:
        NotImplementedError: if no stream for requested physical type is available.

    Returns:
        type[Stream]: jelly stream

    """
    try:
        stream_cls = STREAM_DISPATCH[physical_type]
    except KeyError:
        msg = (
            "no stream class for physical type "
            f"{jelly.PhysicalStreamType.Name(physical_type)}"
        )
        raise NotImplementedError(msg) from None
    return stream_cls


class TripleStream(Stream):
    physical_type = jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    default_delimited_flow_class: ClassVar = FlatTriplesFrameFlow

    def triple(self, terms: Iterable[object]) -> jelly.RdfStreamFrame | None:
        """
        Process one triple to Protobuf messages.

        Note:
            Adds new rows to the current flow and returns StreamFrame if
            frame size conditions are met.

        Args:
            terms (Iterable[object]): RDF terms to encode.

        Returns:
            jelly.RdfStreamFrame | None: stream frame if
                flow supports frames slicing and current flow is full

        """
        new_rows = encode_triple(
            terms,
            term_encoder=self.encoder,
            repeated_terms=self.repeated_terms,
        )
        self.flow.extend(new_rows)
        return self.flow.frame_from_bounds()


class QuadStream(Stream):
    physical_type = jelly.PHYSICAL_STREAM_TYPE_QUADS
    default_delimited_flow_class: ClassVar = FlatQuadsFrameFlow

    def quad(self, terms: Iterable[object]) -> jelly.RdfStreamFrame | None:
        """
        Process one quad to Protobuf messages.

        Args:
            terms (Iterable[object]): terms to encode.

        Returns:
            jelly.RdfStreamFrame | None: stream frame if
                flow supports frames slicing and current flow is full

        """
        new_rows = encode_quad(
            terms,
            term_encoder=self.encoder,
            repeated_terms=self.repeated_terms,
        )
        self.flow.extend(new_rows)
        return self.flow.frame_from_bounds()


class GraphStream(TripleStream):
    physical_type = jelly.PHYSICAL_STREAM_TYPE_GRAPHS
    default_delimited_flow_class: ClassVar = FlatQuadsFrameFlow

    def graph(
        self,
        graph_id: object,
        graph: Iterable[Iterable[object]],
    ) -> Generator[jelly.RdfStreamFrame]:
        """
        Process one graph into a sequence of jelly frames.

        Args:
            graph_id (object): graph id (BN, Literal, iri, default)
            graph (Iterable[Iterable[object]]): iterable of triples (graph's content)

        Yields:
            Generator[jelly.RdfStreamFrame]: jelly frames.

        """
        graph_start = jelly.RdfGraphStart()
        [*graph_rows] = self.encoder.encode_graph(graph_id, graph_start)
        start_row = jelly.RdfStreamRow(graph_start=graph_start)
        graph_rows.append(start_row)
        self.flow.extend(graph_rows)
        for triple in graph:
            if frame := self.triple(triple):  # has frame slicing inside
                yield frame
        end_row = jelly.RdfStreamRow(graph_end=jelly.RdfGraphEnd())
        self.flow.append(end_row)
        if frame := self.flow.frame_from_bounds():
            yield frame


STREAM_DISPATCH: dict[jelly.PhysicalStreamType, type[Stream]] = {
    jelly.PHYSICAL_STREAM_TYPE_TRIPLES: TripleStream,
    jelly.PHYSICAL_STREAM_TYPE_QUADS: QuadStream,
    jelly.PHYSICAL_STREAM_TYPE_GRAPHS: GraphStream,
}



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from unittest.mock import patch

import pytest
from rdflib import Graph, Node, URIRef
from rdflib import Literal as RdfLiteral
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.namespace import RDF
from rdflib.plugins.serializers.nt import _quoteLiteral

from tests.conformance_tests.test_rdf._common import (
    JELLYT,
    MF,
    categorize_by_requires,
)
from tests.meta import RDF_TO_JELLY_MANIFEST, TEST_OUTPUTS_DIR
from tests.serialize import write_generic_sink, write_graph_or_dataset
from tests.utils.rdf_test_cases import jelly_validate, needs_jelly_cli

# Manifest path for to_jelly tests. This is where all test cases are defined.
TO_JELLY_MANIFEST = RDF_TO_JELLY_MANIFEST


@dataclass
class ToJellyTestCase:
    uri: str
    name: str
    action_paths: list[Path]
    options_path: Path | None
    result_path: Path | None
    test_type: str
    category: str
    id: str = field(init=False)

    def __post_init__(self) -> None:
        # Custom id for pytest output to easily see type/category/case
        action_name = (
            self.action_paths[0].parent.name if self.action_paths else "no-action"
        )
        self.id = f"{self.test_type}-{self.category}-{action_name}"


# Load all test cases from the manifest file
# Important: This function translates RDF manifest definitions into Python test cases.
# If new fields appear in manifests (e.g. more metadata), extend logic here.
def load_to_jelly_manifest_cases(manifest_path: Path) -> list[ToJellyTestCase]:
    if not manifest_path.exists():
        return []

    graph = Graph()
    graph.parse(manifest_path, format="turtle")
    manifest_dir = manifest_path.parent
    base_uri = "https://w3id.org/jelly/dev/tests/rdf/to_jelly/"

    test_cases = []
    test_type_map = {
        JELLYT.TestPositive: "positive",
        JELLYT.TestNegative: "negative",
    }

    for test_class, test_type_str in test_type_map.items():
        for test_uri in graph.subjects(RDF.type, test_class):
            if not isinstance(test_uri, URIRef):
                continue

            test_case = _process_test_case(
                graph, test_uri, manifest_dir, base_uri, test_type_str
            )
            if test_case:
                test_cases.append(test_case)

    return test_cases


# Process a single test case entry from the manifest
# This function collects input files (action_paths),
# optional stream options, and expected result.
def _process_test_case(
    graph: Graph,
    test_uri: URIRef,
    manifest_dir: Path,
    base_uri: str,
    test_type_str: str,
) -> ToJellyTestCase | None:
    action_node = graph.value(test_uri, MF.action)
    action_paths, options_path = _process_action_node(
        graph, action_node, manifest_dir, base_uri
    )

    result_path = _process_result_node(graph, test_uri, manifest_dir, base_uri)

    return ToJellyTestCase(
        uri=str(test_uri),
        name=str(graph.value(test_uri, MF.name) or ""),
        action_paths=action_paths,
        options_path=options_path,
        result_path=result_path,
        test_type=test_type_str,
        category=categorize_by_requires(graph, test_uri),
    )


# Process the action node, handling multiple inputs and stream options
# Action can be: a list of input RDF files, a single RDF file, or an options file.
def _process_action_node(
    graph: Graph, action_node: Node | None, manifest_dir: Path, base_uri: str
) -> tuple[list[Path], Path | None]:
    action_paths: list[Path] = []
    options_path = None

    if not action_node:
        return action_paths, options_path

    if (action_node, RDF.first, None) in graph:
        # List of action URIs
        action_uris = graph.items(action_node)
        for action_uri in action_uris:
            uri_str = str(action_uri)
            rel_path = uri_str.replace(base_uri, "")
            if uri_str.endswith("stream_options.jelly"):
                options_path = manifest_dir / rel_path
            else:
                action_paths.append(manifest_dir / rel_path)
    elif str(action_node).endswith("stream_options.jelly"):
        # Single options file
        rel_path = str(action_node).replace(base_uri, "")
        options_path = manifest_dir / rel_path
    else:
        # Single action file
        rel_path = str(action_node).replace(base_uri, "")
        action_paths.append(manifest_dir / rel_path)

    return action_paths, options_path


# Process the result node from the manifest
# Each test case may have one expected output RDF file.
def _process_result_node(
    graph: Graph, test_uri: URIRef, manifest_dir: Path, base_uri: str
) -> Path | None:
    result_node = graph.value(test_uri, MF.result)
    if not result_node:
        return None

    rel_path = str(result_node).replace(base_uri, "")
    return manifest_dir / rel_path


# Patch rdflib nquads serializer to avoid writing default graph id
# Without this patch rdflib includes the default graph id explicitly,
# which breaks comparisons.
def _new_nq_row(triple: tuple[Node, Node, Node], context: Graph) -> str:
    template = "%s " * (3 + (context != DATASET_DEFAULT_GRAPH_ID)) + ".\n"
    args = (
        triple[0].n3(),
        triple[1].n3(),
        _quoteLiteral(triple[2])
        if isinstance(triple[2], RdfLiteral)
        else triple[2].n3(),
        *((context.n3(),) if context != DATASET_DEFAULT_GRAPH_ID else ()),
    )
    return template % args


workaround_rdflib_serializes_default_graph_id = patch(
    "rdflib.plugins.serializers.nquads._nq_row",
    new=_new_nq_row,
)
workaround_rdflib_serializes_default_graph_id.start()

# Collect and categorize test cases for pytest parameterization
ALL_TO_JELLY_CASES = load_to_jelly_manifest_cases(TO_JELLY_MANIFEST)

# Temporary exclusion of a problematic case, can be replaced with xfail later
ALL_TO_JELLY_CASES = [
    case
    for case in ALL_TO_JELLY_CASES
    if not ("pos_014" in case.uri and case.category == "rdf11")
]

RDF11_POSITIVE_CASES = [
    pytest.param(case, id=case.id)
    for case in ALL_TO_JELLY_CASES
    if case.test_type == "positive" and case.category == "rdf11"
]

ALL_POSITIVE_CASES = [
    pytest.param(case, id=case.id)
    for case in ALL_TO_JELLY_CASES
    if case.test_type == "positive"
]

RDF11_NEGATIVE_CASES = [
    pytest.param(case, id=case.id)
    for case in ALL_TO_JELLY_CASES
    if case.test_type == "negative" and case.category == "rdf11"
]

ALL_NEGATIVE_CASES = [
    pytest.param(case, id=case.id)
    for case in ALL_TO_JELLY_CASES
    if case.test_type == "negative"
]


# RDF11 positive tests using to_jelly serialization.
# These ensure our serialization to jelly matches expected RDF 1.1 outputs.
@needs_jelly_cli
@pytest.mark.parametrize("case", RDF11_POSITIVE_CASES)
def test_serializes_rdf11_positive(case: ToJellyTestCase) -> None:
    test_id = case.action_paths[0].parent.name if case.action_paths else "unknown"
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

    input_paths = [p for p in case.action_paths if p.name.startswith("in_")]

    write_graph_or_dataset(
        *[str(p) for p in input_paths],
        options=str(case.options_path) if case.options_path else None,
        out_filename=actual_out,
    )

    # Validate that each input frame matches the corresponding jelly frame
    for frame_no, input_filename in enumerate(input_paths):
        jelly_validate(
            actual_out,
            "--compare-ordered",
            "--compare-frame-indices",
            frame_no,
            "--compare-to-rdf-file",
            input_filename,
            "--options-file",
            str(case.options_path) if case.options_path else "",
            hint=f"Test ID: {test_id}, tested file: {input_filename}",
        )


# Generic integration positive tests.
# Ensure the generic sink produces correct output.
@needs_jelly_cli
@pytest.mark.parametrize("case", ALL_POSITIVE_CASES)
def test_serializes_generic_positive(case: ToJellyTestCase) -> None:
    test_id = case.action_paths[0].parent.name if case.action_paths else "unknown"
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

    write_generic_sink(
        *[str(path) for path in case.action_paths],
        options=str(case.options_path) if case.options_path else None,
        out_filename=actual_out,
    )

    for frame_no, input_filename in enumerate(case.action_paths):
        jelly_validate(
            actual_out,
            "--compare-ordered",
            "--compare-frame-indices",
            frame_no,
            "--compare-to-rdf-file",
            input_filename,
            "--options-file",
            str(case.options_path) if case.options_path else "",
            hint=f"Test ID: {test_id}, tested file: {input_filename}",
        )


# RDF11 negative tests using to_jelly serialization. Parser should fail here.
@needs_jelly_cli
@pytest.mark.parametrize("case", RDF11_NEGATIVE_CASES)
def test_serializing_fails_rdf11_negative(case: ToJellyTestCase) -> None:
    test_id = case.action_paths[0].parent.name if case.action_paths else "unknown"
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

    with pytest.raises(Exception, match=".*"):
        write_graph_or_dataset(
            *[str(path) for path in case.action_paths],
            options=str(case.options_path) if case.options_path else None,
            out_filename=actual_out,
        )


# Generic integration negative tests. Same expectation: serialization must fail.
@needs_jelly_cli
@pytest.mark.parametrize("case", ALL_NEGATIVE_CASES)
def test_serializing_fails_generic_negative(case: ToJellyTestCase) -> None:
    test_id = case.action_paths[0].parent.name if case.action_paths else "unknown"
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

    with pytest.raises(Exception, match=".*"):
        write_generic_sink(
            *[str(path) for path in case.action_paths],
            options=str(case.options_path) if case.options_path else None,
            out_filename=actual_out,
        )



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from enum import IntEnum
from typing import TypeVar, Union
from typing_extensions import TypeAlias

from pyjelly import jelly, options
from pyjelly.errors import JellyConformanceError
from pyjelly.serialize.lookup import LookupEncoder


def split_iri(iri_string: str) -> tuple[str, str]:
    """
    Split iri into prefix and name.

    Args:
        iri_string (str): full iri string.

    Returns:
        tuple[str, str]: iri's prefix and name.

    """
    name = iri_string
    prefix = ""
    for sep in "#", "/":
        prefix, char, name = iri_string.rpartition(sep)
        if char:
            return prefix + char, name
    return prefix, name


T = TypeVar("T")
Rows: TypeAlias = Sequence[jelly.RdfStreamRow]
Statement: TypeAlias = Union[jelly.RdfQuad, jelly.RdfTriple]
HasGraph: TypeAlias = Union[jelly.RdfQuad, jelly.RdfGraphStart]
Terms: TypeAlias = Union[
    jelly.RdfIri, jelly.RdfLiteral, str, jelly.RdfDefaultGraph, jelly.RdfTriple
]


class TermEncoder:
    def __init__(
        self,
        lookup_preset: options.LookupPreset | None = None,
    ) -> None:
        if lookup_preset is None:
            lookup_preset = options.LookupPreset()
        self.lookup_preset = lookup_preset
        self.names = LookupEncoder(lookup_size=lookup_preset.max_names)
        self.prefixes = LookupEncoder(lookup_size=lookup_preset.max_prefixes)
        self.datatypes = LookupEncoder(lookup_size=lookup_preset.max_datatypes)

    def encode_iri_indices(self, iri_string: str) -> tuple[Rows, int, int]:
        """
        Encode lookup indices for IRI.

        Args:
            iri_string (str): full iri in string format.

        Returns:
            tuple[Rows, int, int]: additional rows (if any) and
                indices in prefix and name tables.

        """
        prefix, name = split_iri(iri_string)
        if self.prefixes.lookup.max_size:
            prefix_entry_index = self.prefixes.encode_entry_index(prefix)
        else:
            name = iri_string
            prefix_entry_index = None

        name_entry_index = self.names.encode_entry_index(name)
        term_rows = []

        if prefix_entry_index is not None:
            prefix_entry = jelly.RdfPrefixEntry(id=prefix_entry_index, value=prefix)
            term_rows.append(jelly.RdfStreamRow(prefix=prefix_entry))

        if name_entry_index is not None:
            name_entry = jelly.RdfNameEntry(id=name_entry_index, value=name)
            term_rows.append(jelly.RdfStreamRow(name=name_entry))

        prefix_index = self.prefixes.encode_prefix_term_index(prefix)
        name_index = self.names.encode_name_term_index(name)
        return term_rows, prefix_index, name_index

    def encode_iri(self, iri_string: str, iri: jelly.RdfIri) -> Rows:
        """
        Encode iri.

        Args:
            iri_string (str): full iri in string format.
            iri (jelly.RdfIri): iri to fill

        Returns:
            Rows: extra rows for prefix and name tables, if any.

        """
        term_rows, prefix_index, name_index = self.encode_iri_indices(iri_string)
        iri.prefix_id = prefix_index
        iri.name_id = name_index
        return term_rows

    def encode_default_graph(self, g_default_graph: jelly.RdfDefaultGraph) -> Rows:
        """
        Encode default graph.

        Returns:
            Rows: empty extra rows (for API consistency)

        """
        g_default_graph.CopyFrom(jelly.RdfDefaultGraph())
        return ()

    def encode_literal(
        self,
        *,
        lex: str,
        language: str | None = None,
        datatype: str | None = None,
        literal: jelly.RdfLiteral,
    ) -> Rows:
        """
        Encode literal.

        Args:
            lex (str): lexical form/literal value
            language (str | None, optional): langtag. Defaults to None.
            datatype (str | None, optional): data type if
            it is a typed literal. Defaults to None.
            literal (jelly.RdfLiteral): literal to fill.

        Raises:
            JellyConformanceError: if datatype specified while
                datatable is not used.

        Returns:
            Rows: extra rows (i.e., datatype entries).

        """
        datatype_id = None
        term_rows: tuple[()] | tuple[jelly.RdfStreamRow] = ()

        if datatype and datatype != options.STRING_DATATYPE_IRI:
            if self.datatypes.lookup.max_size == 0:
                msg = (
                    f"can't encode literal with type {datatype}: "
                    "datatype lookup cannot be used if disabled "
                    "(its size was set to 0)"
                )
                raise JellyConformanceError(msg)
            datatype_entry_id = self.datatypes.encode_entry_index(datatype)

            if datatype_entry_id is not None:
                entry = jelly.RdfDatatypeEntry(id=datatype_entry_id, value=datatype)
                term_rows = (jelly.RdfStreamRow(datatype=entry),)

            datatype_id = self.datatypes.encode_datatype_term_index(datatype)

        literal.lex = lex
        if language:
            literal.langtag = language
        if datatype_id:
            literal.datatype = datatype_id
        return term_rows

    def encode_quoted_triple(
        self, terms: Iterable[object], quoted_statement: jelly.RdfTriple
    ) -> Rows:
        """
        Encode a quoted triple.

        Notes:
            Although a triple, it is treated as a part of a statement.
            Repeated terms are not used when encoding quoted triples.

        Args:
            terms (Iterable[object]): triple terms to encode.
            quoted_statement (jelly.RdfTriple): quoted triple to fill.

        Returns:
            Rows: additional stream rows with preceeding
                information (prefixes, names, datatypes rows, if any).

        """
        rows: list[jelly.RdfStreamRow] = []
        terms = iter(terms)
        extra_rows = self.encode_spo(next(terms), Slot.subject, quoted_statement)
        rows.extend(extra_rows)
        extra_rows = self.encode_spo(next(terms), Slot.predicate, quoted_statement)
        rows.extend(extra_rows)
        extra_rows = self.encode_spo(next(terms), Slot.object, quoted_statement)
        rows.extend(extra_rows)
        return rows

    def encode_spo(self, term: object, slot: Slot, statement: Statement) -> Rows:
        msg = f"unsupported term type: {type(term)}"
        raise NotImplementedError(msg)

    def encode_graph(self, term: object, statement: HasGraph) -> Rows:
        msg = f"unsupported term type: {type(term)}"
        raise NotImplementedError(msg)

    def get_iri_field(self, statement: Statement, slot: Slot) -> jelly.RdfIri:
        """Get IRI field directly based on slot."""
        if slot == Slot.subject:
            return statement.s_iri
        if slot == Slot.predicate:
            return statement.p_iri
        return statement.o_iri

    def get_literal_field(self, statement: Statement, slot: Slot) -> jelly.RdfLiteral:
        """Get literal field directly based on slot."""
        if slot == Slot.subject:
            return statement.s_literal
        if slot == Slot.predicate:
            return statement.p_literal
        return statement.o_literal

    def set_bnode_field(
        self, statement: Statement, slot: Slot, identifier: str
    ) -> None:
        """Set bnode field directly based on slot."""
        if slot == Slot.subject:
            statement.s_bnode = identifier
        elif slot == Slot.predicate:
            statement.p_bnode = identifier
        else:
            statement.o_bnode = identifier

    def get_triple_field(self, statement: Statement, slot: Slot) -> jelly.RdfTriple:
        """Get triple term field directly based on slot."""
        if slot == Slot.subject:
            return statement.s_triple_term
        if slot == Slot.predicate:
            return statement.p_triple_term
        return statement.o_triple_term


class Slot(IntEnum):
    subject = 0
    predicate = 1
    object = 2
    graph = 3


def encode_spo(
    terms: Iterator[object],
    term_encoder: TermEncoder,
    repeated_terms: list[object | None],
    statement: Statement,
) -> list[jelly.RdfStreamRow]:
    """
    Encode the s/p/o of a statement.

    Args:
        terms (Iterator[object]): iterator for original terms to encode
        term_encoder (TermEncoder): encoder with lookup tables
        repeated_terms (list[object | None): list of repeated terms.
        statement (Statement): Triple/Quad to fill.

    Returns:
        list[jelly.RdfStreamRow] extra rows to append.

    """
    rows: list[jelly.RdfStreamRow] = []
    s = next(terms)
    if repeated_terms[Slot.subject] != s:
        extra_rows = term_encoder.encode_spo(s, Slot.subject, statement)
        rows.extend(extra_rows)
        repeated_terms[Slot.subject] = s
    p = next(terms)
    if repeated_terms[Slot.predicate] != p:
        extra_rows = term_encoder.encode_spo(p, Slot.predicate, statement)
        rows.extend(extra_rows)
        repeated_terms[Slot.predicate] = p
    o = next(terms)
    if repeated_terms[Slot.object] != o:
        extra_rows = term_encoder.encode_spo(o, Slot.object, statement)
        rows.extend(extra_rows)
        repeated_terms[Slot.object] = o
    return rows


def encode_triple(
    terms: Iterable[object],
    term_encoder: TermEncoder,
    repeated_terms: list[object | None],
) -> list[jelly.RdfStreamRow]:
    """
    Encode one triple.

    Args:
        terms (Iterable[object]): original terms to encode
        term_encoder (TermEncoder): current encoder with lookup tables
        repeated_terms (list[object | None]): list of repeated terms.

    Returns:
        list[jelly.RdfStreamRow]: list of rows to add to the current flow.

    """
    triple = jelly.RdfTriple()
    terms = iter(terms)
    rows = encode_spo(terms, term_encoder, repeated_terms, triple)
    row = jelly.RdfStreamRow(triple=triple)
    rows.append(row)
    return rows


def encode_quad(
    terms: Iterable[object],
    term_encoder: TermEncoder,
    repeated_terms: list[object | None],
) -> list[jelly.RdfStreamRow]:
    """
    Encode one quad.

    Args:
        terms (Iterable[object]): original terms to encode
        term_encoder (TermEncoder): current encoder with lookup tables
        repeated_terms (list[object | None]): list of repeated terms.

    Returns:
        list[jelly.RdfStreamRow]: list of messages to append to current flow.

    """
    terms = iter(terms)
    quad = jelly.RdfQuad()
    rows = encode_spo(terms, term_encoder, repeated_terms, quad)
    g = next(terms)
    if repeated_terms[Slot.graph] != g:
        extra_rows = term_encoder.encode_graph(g, quad)
        rows.extend(extra_rows)
        repeated_terms[Slot.graph] = g
    row = jelly.RdfStreamRow(quad=quad)
    rows.append(row)
    return rows


def encode_namespace_declaration(
    name: str,
    value: str,
    term_encoder: TermEncoder,
) -> list[jelly.RdfStreamRow]:
    """
    Encode namespace declaration.

    Args:
        name (str): namespace prefix label
        value (str): namespace iri
        term_encoder (TermEncoder): current encoder

    Returns:
        list[jelly.RdfStreamRow]: list of messages to append to current flow.

    """
    iri = jelly.RdfIri()
    [*rows] = term_encoder.encode_iri(value, iri=iri)
    declaration = jelly.RdfNamespaceDeclaration(name=name, value=iri)
    row = jelly.RdfStreamRow(namespace=declaration)
    rows.append(row)
    return rows


def encode_options(
    lookup_preset: options.LookupPreset,
    stream_types: options.StreamTypes,
    params: options.StreamParameters,
) -> jelly.RdfStreamRow:
    """
    Encode stream options to ProtoBuf message.

    Args:
        lookup_preset (options.LookupPreset): lookup tables options
        stream_types (options.StreamTypes): physical and logical types
        params (options.StreamParameters): other params.

    Returns:
        jelly.RdfStreamRow: encoded stream options row

    """
    return jelly.RdfStreamRow(
        options=jelly.RdfStreamOptions(
            stream_name=params.stream_name,
            physical_type=stream_types.physical_type,
            generalized_statements=params.generalized_statements,
            rdf_star=params.rdf_star,
            max_name_table_size=lookup_preset.max_names,
            max_prefix_table_size=lookup_preset.max_prefixes,
            max_datatype_table_size=lookup_preset.max_datatypes,
            logical_type=stream_types.logical_type,
            version=params.version,
        )
    )



from pyjelly.integrations.generic.generic_sink import *

# Create a generic sink object
generic_sink = GenericStatementSink()

# Let's add triples one by one
generic_sink.add(
    Triple(
        IRI("http://example.com/subject"),
        IRI("http://example.com/predicate"),
        Literal("Hello", langtag="en"),
    )
)
generic_sink.add(
    Triple(
        BlankNode("B1"),
        IRI("http://example.com/hasName"),
        Literal("Bob"),
    )
)

# Write into a Jelly file
with open("output.jelly", "wb") as out_file:
    generic_sink.serialize(out_file)

print("All done.")



# tests/unit/test_generic_parse.py
from __future__ import annotations

import io
from collections.abc import Iterable, Iterator
from io import BytesIO
from typing import IO, Any

import pytest

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.integrations.generic import parse as gparse
from pyjelly.integrations.generic.generic_sink import (
    IRI,
    Literal,
    Prefix,
    Quad,
    Triple,
)
from pyjelly.integrations.generic.parse import (
    GenericGraphsAdapter,
    GenericStatementSinkAdapter,
    parse_jelly_flat,
    parse_jelly_grouped,
    parse_jelly_to_graph,
)
from pyjelly.options import LookupPreset, StreamParameters, StreamTypes
from pyjelly.parse.decode import ParserOptions


def test_parse_jelly_flat_unsupported_physical_type_raises() -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )
    with pytest.raises(NotImplementedError):
        list(parse_jelly_flat(io.BytesIO(b"data"), frames=iter(()), options=opts))


def test_parse_flat_get_options(monkeypatch: pytest.MonkeyPatch) -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(
            physical_type=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        ),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )

    def dummy_options(_: IO[bytes]) -> tuple[ParserOptions, Iterator[Any]]:
        return opts, iter(())

    monkeypatch.setattr(gparse, "get_options_and_frames", dummy_options)

    out = list(parse_jelly_flat(io.BytesIO(b"x")))
    assert out == []


def test_parse_jelly_grouped_unsupported_physical_type(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )
    monkeypatch.setattr(
        gparse,
        "get_options_and_frames",
        lambda _: (opts, iter(())),
    )
    with pytest.raises(NotImplementedError):
        list(parse_jelly_grouped(BytesIO(b"data")))


def test_namespace_declaration_returns_prefix() -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )
    adapter = GenericStatementSinkAdapter(options=opts)
    p = adapter.namespace_declaration("ex", "http://example.com/")
    assert isinstance(p, Prefix)
    assert p.prefix == "ex"
    assert p.iri == IRI("http://example.com/")


def test_graph_start_end_id() -> None:
    adapter = GenericGraphsAdapter(
        ParserOptions(
            stream_types=StreamTypes(),
            lookup_preset=LookupPreset(),
            params=StreamParameters(),
        )
    )
    adapter.graph_start(IRI("https://g1.com"))
    assert adapter._graph_id == IRI("https://g1.com")
    adapter.graph_end()
    assert adapter._graph_id is None


def test_graph_raises_without_start() -> None:
    adapter = GenericGraphsAdapter(
        ParserOptions(
            stream_types=StreamTypes(),
            lookup_preset=LookupPreset(),
            params=StreamParameters(),
        )
    )
    with pytest.raises(JellyConformanceError):
        _ = adapter.graph


def test_graphs_adapter_triple_appends_graph_id() -> None:
    adapter = GenericGraphsAdapter(
        ParserOptions(
            stream_types=StreamTypes(),
            lookup_preset=LookupPreset(),
            params=StreamParameters(),
        )
    )
    adapter.graph_start(IRI("https://g1.com"))
    s, p, o = IRI("http://s"), IRI("http://p"), IRI("http://o")
    assert adapter.triple([s, p, o]) == Quad(s, p, o, IRI("https://g1.com"))


def test_parse_jelly_grouped_prefixes_triples(monkeypatch: pytest.MonkeyPatch) -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(
            physical_type=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        ),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )
    monkeypatch.setattr(
        gparse,
        "get_options_and_frames",
        lambda _inp: (opts, iter(())),
    )

    def dummy_triples_stream(
        frames: Iterable[jelly.RdfStreamFrame],  # noqa: ARG001
        options: ParserOptions,  # noqa: ARG001
    ) -> Any:
        yield [
            Prefix("ex", IRI("http://example.com/")),
            Triple(
                IRI("http://example.com/s"),
                IRI("http://example.com/p"),
                Literal("http://example.com/o"),
            ),
        ]

    monkeypatch.setattr(gparse, "parse_triples_stream", dummy_triples_stream)
    sink = list(parse_jelly_grouped(io.BytesIO(b"data")))

    assert len(sink) == 1
    assert any(isinstance(x, Triple) for x in sink[0].store)


def test_parse_jelly_grouped_prefixes_quads(monkeypatch: pytest.MonkeyPatch) -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(
            physical_type=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
        ),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )
    monkeypatch.setattr(
        gparse,
        "get_options_and_frames",
        lambda _inp: (opts, iter(())),
    )

    def fake_parse_quads_stream(
        frames: Iterable[jelly.RdfStreamFrame],  # noqa: ARG001
        options: ParserOptions,  # noqa: ARG001
    ) -> Any:
        yield [
            Prefix("ex", IRI("http://example.com/")),
            Quad(
                IRI("http://example.com/s"),
                IRI("http://example.com/p"),
                IRI("http://example.com/o"),
                IRI("http://example.com/g"),
            ),
        ]

    monkeypatch.setattr(gparse, "parse_quads_stream", fake_parse_quads_stream)

    sink = list(parse_jelly_grouped(io.BytesIO(b"dummy")))

    assert len(sink) == 1
    assert any(isinstance(x, Quad) for x in sink[0].store)


def test_parse_jelly_to_graph_prefix(monkeypatch: pytest.MonkeyPatch) -> None:
    opts = ParserOptions(
        stream_types=StreamTypes(),
        lookup_preset=LookupPreset(),
        params=StreamParameters(),
    )
    monkeypatch.setattr(gparse, "get_options_and_frames", lambda _: (opts, iter(())))

    def dummy_parse_jelly_flat(
        *,
        inp: IO[bytes],  # noqa: ARG001
        frames: Iterable[jelly.RdfStreamFrame],  # noqa: ARG001
        options: ParserOptions,  # noqa: ARG001
    ) -> Any:
        yield Prefix("ex", IRI("http://example.com/"))
        yield Triple(
            IRI("http://example.com/s"),
            IRI("http://example.com/p"),
            Literal("http://example.com/o"),
        )

    monkeypatch.setattr(gparse, "parse_jelly_flat", dummy_parse_jelly_flat)

    sink = parse_jelly_to_graph(io.BytesIO(b"dummy"))
    assert any(isinstance(st, Triple) for st in sink.store)



import io

from pyjelly.integrations.generic.generic_sink import *

g1 = GenericStatementSink()

g1.add(
    Triple(
        IRI("http://example.com/subject"),
        IRI("http://example.com/predicate"),
        Literal("Hello", langtag="en"),
    )
)

# Write the data into a byte buffer (bytes type)
with io.BytesIO() as write_buffer:
    g1.serialize(write_buffer)
    data = write_buffer.getvalue()

print(f"Serialized data size: {len(data)} bytes")

# Parse the data back
g2 = GenericStatementSink()
with io.BytesIO(data) as read_buffer:
    g2.parse(read_buffer)

print("\nParsed triples:")
for statement in g2:
    print(statement)



import pytest

from pyjelly.integrations.generic.generic_sink import (
    GenericStatementSink,
    Literal,
    Triple,
)
from tests.utils.generic_sink_test_parser import GenericSinkParser


@pytest.mark.parametrize(
    (
        "lex",
        "langtag",
        "datatype",
        "expected_str",
    ),
    [
        ("hello", None, None, '"hello"'),
        ("", None, None, '""'),
        ("123", None, None, '"123"'),
        ("true", None, None, '"true"'),
        ("hello", "en", None, '"hello"@en'),
        ("bonjour", "fr", None, '"bonjour"@fr'),
        ("hallo", "de", None, '"hallo"@de'),
        ("hello", "en-US", None, '"hello"@en-US'),
        ("café", "fr", None, '"café"@fr'),
        (
            "123",
            None,
            "http://www.w3.org/2001/XMLSchema#integer",
            '"123"^^<http://www.w3.org/2001/XMLSchema#integer>',
        ),
        (
            "3.14",
            None,
            "http://www.w3.org/2001/XMLSchema#float",
            '"3.14"^^<http://www.w3.org/2001/XMLSchema#float>',
        ),
        (
            "true",
            None,
            "http://www.w3.org/2001/XMLSchema#boolean",
            '"true"^^<http://www.w3.org/2001/XMLSchema#boolean>',
        ),
        (
            "2023-01-01",
            None,
            "http://www.w3.org/2001/XMLSchema#date",
            '"2023-01-01"^^<http://www.w3.org/2001/XMLSchema#date>',
        ),
        (
            "hello",
            None,
            "http://www.w3.org/2001/XMLSchema#string",
            '"hello"^^<http://www.w3.org/2001/XMLSchema#string>',
        ),
        ("line1\nline2", None, None, '"line1\nline2"'),
        ("tab\there", None, None, '"tab\there"'),
        ("backslash\\here", None, None, '"backslash\\here"'),
        ("résumé", "fr", None, '"résumé"@fr'),
        ("München", "de", None, '"München"@de'),
        ("日本語", "ja", None, '"日本語"@ja'),
        ("🚀", None, None, '"🚀"'),
        (" ", None, None, '" "'),
        ("  whitespace  ", None, None, '"  whitespace  "'),
        ("", "en", None, '""@en'),
        (
            "",
            None,
            "http://www.w3.org/2001/XMLSchema#string",
            '""^^<http://www.w3.org/2001/XMLSchema#string>',
        ),
    ],
)
def test_literal_parsing(
    lex: str, langtag: str, datatype: str, expected_str: str
) -> None:
    """Test literals parsing."""
    literal = str(Literal(lex, langtag, datatype))
    test_statement = f"_:bn <http://www.examples.org/predicate> {literal} ."
    gp = GenericSinkParser(GenericStatementSink())
    triple = gp.parse_statement(test_statement, Triple)
    assert str(triple[2]) == expected_str



from typing import Callable, Union, cast

import pytest
from rdflib import Graph
from rdflib.graph import Dataset

from pyjelly import jelly
from pyjelly.errors import JellyAssertionError
from pyjelly.integrations.rdflib.serialize import guess_options, guess_stream
from pyjelly.options import StreamParameters
from pyjelly.serialize.streams import QuadStream, SerializerOptions, TripleStream

Store = Union[Graph, Dataset]


@pytest.mark.parametrize(
    ("make_store", "expected_logical"),
    [
        (Graph, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (Dataset, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
    ],
    ids=["graph->triples", "dataset->quads"],
)
def test_defaults_rdflib(
    make_store: Callable[[], Store],
    expected_logical: int,
) -> None:
    store = make_store()
    opts = guess_options(store)
    assert opts.logical_type == expected_logical
    assert opts.params.rdf_star is False
    assert opts.params.generalized_statements is False


@pytest.mark.parametrize(
    ("make_store", "logical", "expected_physical"),
    [
        (
            Graph,
            jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            TripleStream,
        ),
        (
            Dataset,
            jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            QuadStream,
        ),
    ],
    ids=["graph+triples", "dataset+quads"],
)
def test_override_rdflib_compatible(
    make_store: Callable[[], Store],
    logical: int,
    expected_physical: type,
) -> None:
    store = make_store()
    user_opts = SerializerOptions(
        logical_type=cast(jelly.LogicalStreamType, logical),
        params=StreamParameters(
            rdf_star=True,
            generalized_statements=True,
        ),
    )
    stream = guess_stream(user_opts, store)

    assert isinstance(stream, expected_physical)
    assert stream.options.logical_type == logical
    assert stream.options.params.rdf_star is True
    assert stream.options.params.generalized_statements is True


@pytest.mark.parametrize(
    ("make_store", "logical"),
    [
        (Graph, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (Dataset, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
    ],
    ids=["graph+quads-error", "dataset+triples-error"],
)
def test_override_rdflib_incompatible(
    make_store: Callable[[], Store],
    logical: int,
) -> None:
    store = make_store()
    user_opts = SerializerOptions(
        logical_type=cast(jelly.LogicalStreamType, logical),
        params=StreamParameters(
            rdf_star=True,
            generalized_statements=True,
        ),
    )
    with pytest.raises(
        JellyAssertionError,
        match="is not compatible with",
    ):
        guess_stream(user_opts, store)



from __future__ import annotations

import tempfile
from dataclasses import replace
from functools import partial
from itertools import chain, product
from pathlib import Path

import pytest

from pyjelly.options import LookupPreset
from tests.e2e_tests.ser_des.base_ser_des import BaseSerDes
from tests.e2e_tests.ser_des.generic_ser_des import GenericSerDes
from tests.e2e_tests.ser_des.rdflib_ser_des import RdflibSerDes
from tests.utils.rdf_test_cases import jelly_cli, needs_jelly_cli

DEFAULT_PRESET = replace(LookupPreset.small(), max_prefixes=0)
DEFAULT_FRAME_SIZE = 200

jelly_validate = partial(jelly_cli, "rdf", "validate", "--compare-ordered")
jelly_from_jelly = partial(jelly_cli, "rdf", "from-jelly")
rdf_to_jelly = partial(jelly_cli, "rdf", "to-jelly")


class End2EndOptionSetup:
    """Set up stream options, file size and file name for rdf E2E tests."""

    test_root: Path = Path("tests/e2e_test_cases/")

    def setup_ser_des(self) -> list[tuple[BaseSerDes, BaseSerDes, LookupPreset, int]]:
        """Set up test serializer, deserializer, options and frame_size."""
        ser = [RdflibSerDes()]
        des = [RdflibSerDes()]
        # We want to have a variety of options to test
        # Particularly examples of small lookup sizes
        # and a lack of prefix
        small = LookupPreset.small()
        no_prefixes = replace(LookupPreset.small(), max_prefixes=0)
        tiny_lookups = replace(LookupPreset.small(), max_names=16, max_prefixes=8)
        big = LookupPreset()
        presets = [small, no_prefixes, tiny_lookups, big]
        frame_sizes = [1, 4, 200, 10_000]
        return list(product(ser, des, presets, frame_sizes))

    def setup_triple_files(
        self,
    ) -> list[tuple[BaseSerDes, BaseSerDes, LookupPreset, int, Path]]:
        """Set up options for each of the test triple files."""
        test_dir: Path = self.test_root / "triples_rdf_1_1"
        files = test_dir.glob("*.nt")
        options = self.setup_ser_des()
        return list(chain(*[[(*o, f) for o in options] for f in files]))

    def setup_quad_files(
        self,
    ) -> list[tuple[BaseSerDes, BaseSerDes, LookupPreset, int, Path]]:
        """Set up options for each of the test quad files."""
        test_dir: Path = self.test_root / "quads_rdf_1_1"
        files = test_dir.glob("*.nq")
        options = self.setup_ser_des()
        return list(chain(*[[(*o, f) for o in options] for f in files]))


class End2EndOptionSetupGeneric:
    """Set up stream options, file size and file name for generic sink E2E tests."""

    test_root: Path = Path("tests/e2e_test_cases/")

    def setup_ser_des(
        self,
    ) -> list[tuple[GenericSerDes, GenericSerDes, LookupPreset, int]]:
        ser = [GenericSerDes()]
        des = [GenericSerDes()]
        small = LookupPreset.small()
        no_prefixes = replace(LookupPreset.small(), max_prefixes=0)
        tiny_lookups = replace(LookupPreset.small(), max_names=16, max_prefixes=8)
        big = LookupPreset()
        presets = [small, no_prefixes, tiny_lookups, big]
        frame_sizes = [1, 4, 200, 10_000]
        return list(product(ser, des, presets, frame_sizes))

    def setup_triple_files(
        self,
    ) -> list[tuple[GenericSerDes, GenericSerDes, LookupPreset, int, Path]]:
        test_dir: Path = self.test_root / "triples_rdf_1_1"
        files = test_dir.glob("*.nt")
        options = self.setup_ser_des()
        return list(chain(*[[(*o, f) for o in options] for f in files]))

    def setup_quad_files(
        self,
    ) -> list[tuple[GenericSerDes, GenericSerDes, LookupPreset, int, Path]]:
        test_dir: Path = self.test_root / "quads_rdf_1_1"
        files = test_dir.glob("*.nq")
        options = self.setup_ser_des()
        return list(chain(*[[(*o, f) for o in options] for f in files]))


class End2EndOptionSetupCross:
    """Set up stream options, file size and file name for cross E2E tests."""

    rdf_root: Path = Path("tests/e2e_test_cases/")

    def triples(self) -> list[Path]:
        return sorted((self.rdf_root / "triples_rdf_1_1").glob("*.nt"))

    def quads(self) -> list[Path]:
        return sorted((self.rdf_root / "quads_rdf_1_1").glob("*.nq"))


class TestEnd2End:
    setup = End2EndOptionSetup()

    @pytest.mark.parametrize(
        ("ser", "des", "preset", "frame_size", "file"), setup.setup_triple_files()
    )
    def test_triple_files(
        self,
        ser: BaseSerDes,
        des: BaseSerDes,
        preset: LookupPreset,
        frame_size: int,
        file: Path,
    ) -> None:
        nt_reader = RdflibSerDes()
        with file.open("rb") as f:
            triples = nt_reader.read_triples(f.read())
            jelly_io = ser.write_triples_jelly(triples, preset, frame_size)
            new_g = des.read_triples_jelly(jelly_io)
            assert set(triples) == set(new_g)

    @pytest.mark.parametrize(
        ("ser", "des", "preset", "frame_size", "file"), setup.setup_quad_files()
    )
    def test_quad_files(
        self,
        ser: BaseSerDes,
        des: BaseSerDes,
        preset: LookupPreset,
        frame_size: int,
        file: Path,
    ) -> None:
        nq_reader = RdflibSerDes()
        with file.open("rb") as f:
            quads = nq_reader.read_quads(f.read())
            jelly_io = ser.write_quads_jelly(quads, preset, frame_size)
            new_g = des.read_quads_jelly(jelly_io)
            assert set(quads) == set(new_g)


class TestEnd2EndGeneric:
    setup = End2EndOptionSetupGeneric()

    @needs_jelly_cli
    @pytest.mark.parametrize(
        ("ser", "des", "preset", "frame_size", "file"), setup.setup_triple_files()
    )
    def test_triple_files(
        self,
        ser: GenericSerDes,
        des: GenericSerDes,
        preset: LookupPreset,
        frame_size: int,
        file: Path,
    ) -> None:
        reader = GenericSerDes()
        gen_input_jelly = rdf_to_jelly(file)
        triples = reader.read_triples(gen_input_jelly)
        jelly = ser.write_triples_jelly(triples, preset, frame_size)
        new_g = des.read_triples(jelly)
        assert set(triples) == set(new_g)

    @needs_jelly_cli
    @pytest.mark.parametrize(
        ("ser", "des", "preset", "frame_size", "file"), setup.setup_quad_files()
    )
    def test_quad_files(
        self,
        ser: GenericSerDes,
        des: GenericSerDes,
        preset: LookupPreset,
        frame_size: int,
        file: Path,
    ) -> None:
        reader = GenericSerDes()
        gen_input_jelly = rdf_to_jelly(file)
        quads = reader.read_quads(gen_input_jelly)
        jelly = ser.write_quads_jelly(quads, preset, frame_size)
        new_g = des.read_quads(jelly)
        assert set(quads) == set(new_g)


class TestEnd2EndCross:
    setup = End2EndOptionSetupCross()
    triples_cases = setup.triples()
    quads_cases = setup.quads()

    @needs_jelly_cli
    @pytest.mark.parametrize(
        "r_path",
        triples_cases,
        ids=[f"triples:{p.name}" for p in triples_cases],
    )
    def test_cross_generic_rdf_triple(self, r_path: Path) -> None:
        gen = GenericSerDes()
        rdf = RdflibSerDes()
        gen_input_jelly = rdf_to_jelly(r_path)
        g_triples = gen.read_triples(gen_input_jelly)
        g_jelly = gen.write_triples(g_triples)
        with r_path.open("rb") as file_rdf:
            r_triples = rdf.read_triples(file_rdf.read())
            r_jelly = rdf.write_triples_jelly(
                r_triples, DEFAULT_PRESET, DEFAULT_FRAME_SIZE
            )
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_gen = Path(temp_dir) / "gen.jelly"
            temp_rdf = Path(temp_dir) / "rdf.jelly"
            temp_gen.write_bytes(g_jelly)
            temp_rdf.write_bytes(r_jelly)
            jelly_validate(temp_gen, "--compare-to-rdf-file", temp_rdf)

    @needs_jelly_cli
    @pytest.mark.parametrize(
        "r_path",
        quads_cases,
        ids=[f"quads:{p.name}" for p in quads_cases],
    )
    def test_cross_generic_rdf_quad(self, r_path: Path) -> None:
        gen = GenericSerDes()
        rdf = RdflibSerDes()
        gen_input_jelly = rdf_to_jelly(r_path)
        g_quads = gen.read_quads(gen_input_jelly)
        g_jelly = gen.write_quads(g_quads)
        with r_path.open("rb") as file_rdf:
            r_quads = rdf.read_quads(file_rdf.read())
            r_jelly = rdf.write_quads_jelly(r_quads, DEFAULT_PRESET, DEFAULT_FRAME_SIZE)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_gen = Path(temp_dir) / "gen.jelly"
            temp_rdf = Path(temp_dir) / "rdf.jelly"
            temp_gen.write_bytes(g_jelly)
            temp_rdf.write_bytes(r_jelly)
            jelly_validate(temp_gen, "--compare-to-rdf-file", temp_rdf)



from __future__ import annotations

import io
from contextlib import nullcontext
from dataclasses import dataclass
from typing import Any, Callable

import pytest

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.integrations.generic.generic_sink import (
    IRI,
    DefaultGraph,
    GenericStatementSink,
    Literal,
    Quad,
    Triple,
)
from pyjelly.integrations.generic.parse import (
    parse_jelly_flat as generic_parse_jelly_flat,
)
from pyjelly.integrations.generic.parse import (
    parse_jelly_grouped as generic_parse_jelly_grouped,
)
from pyjelly.integrations.generic.serialize import (
    flat_stream_to_file,
    graphs_stream_frames,
    grouped_stream_to_file,
)
from pyjelly.options import StreamParameters
from pyjelly.parse.ioutils import get_options_and_frames
from pyjelly.serialize.streams import SerializerOptions


def _make_flat_triples_bytes() -> bytes:
    sink = GenericStatementSink()

    subject = IRI("http://example.org/subject")
    predicate = IRI("http://example.org/predicate")
    object1 = Literal("value1")
    object2 = IRI("http://example.org/object")

    sink.add(Triple(subject, predicate, object1))
    sink.add(Triple(subject, predicate, object2))

    output = io.BytesIO()
    options = SerializerOptions(
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        params=StreamParameters(
            generalized_statements=True,
            rdf_star=True,
            namespace_declarations=False,
        ),
    )
    flat_stream_to_file(sink.store, output, options=options)
    return output.getvalue()


def _make_flat_quads_bytes() -> bytes:
    sink = GenericStatementSink()

    subject = IRI("http://example.org/subject")
    predicate = IRI("http://example.org/predicate")
    object1 = Literal("value1")
    graph = IRI("http://example.org/graph")

    sink.add(Quad(subject, predicate, object1, graph))
    sink.add(Quad(subject, predicate, object1, DefaultGraph))

    output = io.BytesIO()
    options = SerializerOptions(
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
        params=StreamParameters(
            generalized_statements=True,
            rdf_star=True,
            namespace_declarations=False,
        ),
    )
    flat_stream_to_file(sink.store, output, options=options)
    return output.getvalue()


def _make_physical_graphs_bytes() -> bytes:
    from pyjelly.integrations.generic.serialize import GenericSinkTermEncoder
    from pyjelly.serialize.streams import GraphStream

    options = SerializerOptions(
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
        params=StreamParameters(
            generalized_statements=True,
            rdf_star=True,
            namespace_declarations=False,
        ),
    )

    stream = GraphStream(
        encoder=GenericSinkTermEncoder(),
        options=options,
    )

    sink = GenericStatementSink()

    sink.add(
        Quad(
            IRI("http://ex/s1"),
            IRI("http://ex/p1"),
            IRI("http://ex/o1"),
            IRI("http://ex/g1"),
        )
    )
    sink.add(
        Quad(
            IRI("http://ex/s2"),
            IRI("http://ex/p2"),
            IRI("http://ex/o2"),
            IRI("http://ex/g2"),
        )
    )

    output = io.BytesIO()

    frames = list(graphs_stream_frames(stream, sink))

    for frame in frames:
        output.write(frame.SerializeToString())

    return output.getvalue()


def _make_grouped_graphs_bytes() -> bytes:
    graph1 = GenericStatementSink(IRI("http://example.org/graph1"))
    graph2 = GenericStatementSink(IRI("http://example.org/graph2"))

    subject = IRI("http://example.org/subject")
    predicate = IRI("http://example.org/predicate")

    graph1.add(Triple(subject, predicate, Literal("graph1_value")))
    graph2.add(Triple(subject, predicate, Literal("graph2_value")))

    output = io.BytesIO()
    options = SerializerOptions(
        logical_type=jelly.LOGICAL_STREAM_TYPE_GRAPHS,
        params=StreamParameters(
            generalized_statements=True,
            rdf_star=True,
            namespace_declarations=False,
        ),
    )
    grouped_stream_to_file((g for g in [graph1, graph2]), output, options=options)
    return output.getvalue()


@dataclass(frozen=True)
class Case:
    make_bytes: Callable[[], bytes]
    expected_physical: int
    expected_logical: int
    parser: Callable[..., Any]
    strict: bool
    should_raise: bool
    msg: str | None


@pytest.mark.parametrize(
    "case",
    [
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=generic_parse_jelly_grouped,
            strict=True,
            should_raise=True,
            msg="expected GROUPED",
        ),
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=generic_parse_jelly_grouped,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=generic_parse_jelly_flat,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_flat,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_grouped_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_GRAPHS,
            parser=generic_parse_jelly_flat,
            strict=True,
            should_raise=True,
            msg="expected FLAT",
        ),
        Case(
            make_bytes=_make_grouped_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_GRAPHS,
            parser=generic_parse_jelly_grouped,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_grouped,
            strict=True,
            should_raise=True,
            msg="expected GROUPED",
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_grouped,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_quads_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_QUADS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_flat,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_flat_triples_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            parser=generic_parse_jelly_flat,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_grouped_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_GRAPHS,
            parser=generic_parse_jelly_grouped,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_flat,
            strict=True,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_flat,
            strict=False,
            should_raise=False,
            msg=None,
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_grouped,
            strict=True,
            should_raise=True,
            msg="expected GROUPED",
        ),
        Case(
            make_bytes=_make_physical_graphs_bytes,
            expected_physical=jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            expected_logical=jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            parser=generic_parse_jelly_grouped,
            strict=False,
            should_raise=False,
            msg=None,
        ),
    ],
)
def test_generic_logical_matrix(case: Case) -> None:
    data = case.make_bytes()
    opts, _ = get_options_and_frames(io.BytesIO(data))
    assert opts.stream_types.physical_type == case.expected_physical
    assert opts.stream_types.logical_type == case.expected_logical

    ctx = (
        pytest.raises(JellyConformanceError, match=case.msg)
        if case.should_raise
        else nullcontext()
    )
    with ctx:
        list(case.parser(io.BytesIO(data), logical_type_strict=case.strict))


def test_generic_flat_strict_requires_stream_types() -> None:
    dummy = b"x"
    options = type("Opt", (), {"stream_types": None})()
    frames: list[object] = []
    from unittest.mock import patch

    with (
        patch(
            "pyjelly.integrations.generic.parse.get_options_and_frames",
            return_value=(options, frames),
        ),
        pytest.raises(JellyConformanceError, match="requires options.stream_types"),
    ):
        list(generic_parse_jelly_flat(io.BytesIO(dummy), logical_type_strict=True))


def test_generic_grouped_strict_unspecified_raises() -> None:
    class ST:
        flat = False
        logical_type = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
        physical_type = jelly.PHYSICAL_STREAM_TYPE_TRIPLES

    dummy = b"x"
    options = type("Opt", (), {"stream_types": ST()})()
    frames: list[object] = []
    from unittest.mock import patch

    with (
        patch(
            "pyjelly.integrations.generic.parse.get_options_and_frames",
            return_value=(options, frames),
        ),
        pytest.raises(JellyConformanceError, match="expected GROUPED"),
    ):
        list(generic_parse_jelly_grouped(io.BytesIO(dummy), logical_type_strict=True))


def test_generic_flat_unsupported_physical_raises() -> None:
    class ST:
        flat = True
        logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
        physical_type = jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED

    dummy = b"x"
    options = type("Opt", (), {"stream_types": ST()})()
    frames: list[object] = []
    from unittest.mock import patch

    with (
        patch(
            "pyjelly.integrations.generic.parse.get_options_and_frames",
            return_value=(options, frames),
        ),
        pytest.raises(NotImplementedError, match="not supported"),
    ):
        list(generic_parse_jelly_flat(io.BytesIO(dummy), logical_type_strict=False))



import importlib
from pathlib import Path

from rdflib import Graph, Literal, Namespace, URIRef

rdflib_serialize = importlib.import_module("pyjelly.integrations.rdflib.serialize")


def test_rdflib_roundtrip_keeps_prefixes(tmp_path: Path) -> None:
    g = Graph()
    ex_ns = Namespace("http://example.org/")
    g.namespace_manager.bind("ex", ex_ns)
    g.add((ex_ns.alice, URIRef("http://xmlns.com/foaf/0.1/name"), Literal("Alice")))

    options = rdflib_serialize.SerializerOptions(
        params=rdflib_serialize.StreamParameters(
            generalized_statements=False,
            rdf_star=False,
            namespace_declarations=True,
        )
    )

    out = tmp_path / "g.jelly"
    g.serialize(out.as_posix(), format="jelly", options=options)

    g2 = Graph()
    g2.parse(out.as_posix(), format="jelly")

    ns = dict(g2.namespaces())
    assert "ex" in ns
    assert str(ns["ex"]) == "http://example.org/"
